from pydantic import Field, field_validator
from quart import Blueprint, g
from quart_schema import validate_request, validate_response
from sqlalchemy import select

from foxnouns import tasks
from foxnouns.auth import require_auth
from foxnouns.db import User
from foxnouns.db.aio import async_session
from foxnouns.db.snowflake import Snowflake
from foxnouns.db.util import create_token, generate_token, is_self, user_from_ref
from foxnouns.exceptions import ErrorCode, NotFoundError
from foxnouns.models import BasePatchModel
from foxnouns.models.user import SelfUserModel, UserModel, check_username
from foxnouns.settings import BASE_DOMAIN

bp = Blueprint("users_v2", __name__)


@bp.get("/api/v2/users/<user_ref>", host=BASE_DOMAIN)
@validate_response(UserModel, 200)
async def get_user(user_ref: str):
    async with async_session() as session:
        user = await user_from_ref(session, user_ref)
        if not user:
            raise NotFoundError("User not found", type=ErrorCode.UserNotFound)

        return (
            SelfUserModel.model_validate(user)
            if is_self(user)
            else UserModel.model_validate(user)
        )


class EditUserRequest(BasePatchModel):
    username: str | None = Field(
        min_length=2, max_length=40, pattern=r"^[\w\-\.]{2,40}$", default=None
    )
    display_name: str | None = Field(max_length=100, default=None)
    bio: str | None = Field(max_length=1024, default=None)

    avatar: str | None = Field(max_length=1_000_000, default=None)

    @field_validator("username")
    @classmethod
    def check_username(cls, value):
        return check_username(value)


@bp.patch("/api/v2/users/@me", host=BASE_DOMAIN)
@require_auth(scope="user.update")
@validate_request(EditUserRequest)
@validate_response(SelfUserModel, 200)
async def edit_user(data: EditUserRequest):
    """Updates the current user."""

    async with async_session() as session:
        user = await session.scalar(select(User).where(User.id == g.user.id))
        await user.awaitable_attrs.members

        if data.username:
            user.username = data.username
        if data.is_set("display_name"):
            user.display_name = data.display_name
        if data.is_set("bio"):
            user.bio = data.bio

        await session.commit()

        if data.is_set("avatar"):
            if data.avatar:
                tasks.process_user_avatar.delay(user.id, data.avatar)
            else:
                tasks.delete_user_avatar.delay(user.id)

        return SelfUserModel.model_validate(user)


class DebugUserData(BasePatchModel):
    username: str


class DebugUserResponse(SelfUserModel):
    token: str


@bp.post("/api/v2/users/debug", host=BASE_DOMAIN)
@validate_request(DebugUserData)
@validate_response(DebugUserResponse, 200)
async def debug_create_user(data: DebugUserData):
    """Creates a user from just a username, and returns it along with a token.
    FIXME: this must be removed **BEFORE** deploying to production (or even public testing)
    """

    async with async_session() as session:
        user = User(id=Snowflake.generate_int(), username=data.username)
        await session.commit()

        session.add(user)
        token = await create_token(session, user, ["*"])
        await session.commit()
        await user.awaitable_attrs.members

        user.token = generate_token(token)
        return DebugUserResponse.model_validate(user)
